import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import NeuralAES

from .utils import get_linear_mapping, MUL2, MUL3, SBOX

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 # torch.float16 for benchmarking

class TTablesNeuralAES(NeuralAES):
    def __init__(self, secret_key, direction='Encryption', c_parameter=1.0, number_of_rounds=10, protected=False,
                 epsilon=None):
        '''
        Builds a number_of_rounds instance of the AES embedded with the key secret_key, given as an integer. The corresponding NN uses c_parameter in the corner functions.
        If protected is set to True, the protection measure described in the paper is applied, with the specified epsilon value.
        '''
        super().__init__(secret_key, direction, c_parameter, number_of_rounds, protected, epsilon)

        SM2 = np.uint8(MUL2)[np.uint8(SBOX)]
        SM3 = np.uint8(MUL3)[np.uint8(SBOX)]

        if direction == 'Encryption':
            self.m2w_s = torch.tensor(get_linear_mapping(SM2), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m3w_s = torch.tensor(get_linear_mapping(SM3), dtype=DTYPE, requires_grad=False).to(DEVICE)
        # TTables is only applicable in the encryption direction, so the performance gain does not port to decryption
        
    def SB_MC_fused(self, state):
        '''
        Fused SubBytes + MixColumns for non-final encryption rounds.
        Expects ShiftRows to have already been applied.
        '''
        all_corners = F.relu(F.linear(state, self.byte_op_w, self.byte_op_b))
        SB = F.linear(all_corners, self.sbw_s)
        M2 = F.linear(all_corners, self.m2w_s)
        M3 = F.linear(all_corners, self.m3w_s)
        X = torch.stack((M2, 
                         torch.cat((M3[:, 1:], M3[:, :1]), dim=1),
                         torch.cat((SB[:, 2:], SB[:, :2]), dim=1),
                         torch.cat((SB[:, 3:], SB[:, :3]), dim=1)
                         ), dim=-1)
        return self.apply_op(X, self.xor4w, self.xor4b)
    
    def AES_round(self, state, final=False):
        X = self.SR(state)
        if not final:
            X = self.SB_MC_fused(X)
        else:
            X = self.SB(X)
        return X


import numpy as np
import torch
import torch.nn.functional as F

from nnaes.base import NeuralAES, SumBumpsLayer, DEVICE, DTYPE
from nnaes.ttables import TTablesNeuralAES
from nnaes.utils import integer_to_bitvector, state_as_ints

def encrypt_list_of_plaintexts(plaintexts, model, dtype=DTYPE):
    state = [integer_to_bitvector(x) for x in plaintexts]
    model = model.to(DEVICE)
    model.eval()
    state_tensor = torch.tensor(state, dtype=dtype).to(DEVICE)
    with torch.inference_mode():
        state_tensor = model(state_tensor)
    return state_as_ints(state_tensor)


def test_aes_testvectors(protected=False, epsilon=None):
    p_fips = [0x3243f6a8885a308d313198a2e0370734]
    k_fips = 0x2b7e151628aed2a6abf7158809cf4f3c
    c_fips = [0x3925841d02dc09fbdc118597196a0b32]

    k_aesavs = 0
    p_aesavs = [
        0x80000000000000000000000000000000,
        0xc0000000000000000000000000000000,
        0xe0000000000000000000000000000000,
        0xf0000000000000000000000000000000,
        0xf8000000000000000000000000000000,
    ]
    c_aesavs = [
        0x3ad78e726c1ec02b7ebfe92b23d9ec34,
        0xaae5939c8efdf2f04e60b9fe7117b2c2,
        0xf031d4d74f5dcbf39daaf8ca3af6e527,
        0x96d9fd5cc4f07441727df0f33e401a36,
        0x30ccdb044646d7e1f3ccea3dca08b8c0,
    ]

    for model_type in [NeuralAES, TTablesNeuralAES]:
        for c_parameter in [1.0, 0.5]:
            aes_enc = model_type(k_fips, c_parameter=c_parameter, protected=protected, epsilon=epsilon)
            ct = encrypt_list_of_plaintexts(p_fips, aes_enc)
            assert ct[0] == c_fips[0]

            aes_enc = model_type(k_aesavs, c_parameter=c_parameter, protected=protected, epsilon=epsilon)
            ct = encrypt_list_of_plaintexts(p_aesavs, aes_enc)
            for i in range(5):
                assert ct[i] == c_aesavs[i]

        for c_parameter in [1.0, 0.5]:
            aes_dec = model_type(
                k_fips,
                direction="Decryption",
                c_parameter=c_parameter,
                protected=protected,
                epsilon=epsilon,
            )
            pt = encrypt_list_of_plaintexts(c_fips, aes_dec)
            assert pt[0] == p_fips[0]

            aes_dec = model_type(
                k_aesavs,
                direction="Decryption",
                c_parameter=c_parameter,
                protected=protected,
                epsilon=epsilon,
            )
            pt = encrypt_list_of_plaintexts(c_aesavs, aes_dec)
            for i in range(5):
                assert pt[i] == p_aesavs[i]


def test_unprotected_aes_testvectors():
    test_aes_testvectors(protected=False)


def test_protected_aes_testvectors():
    test_aes_testvectors(protected=True, epsilon=1 / 4)


def test_protected_aes_functionality(num_samples=10**3, epsilon=1 / 4):
    for model_type in [NeuralAES, TTablesNeuralAES]:
        torch.cuda.empty_cache()
        k = 0x2b7e151628aed2a6abf7158809cf4f3c
        model = model_type(k, c_parameter=0.5, protected=True, epsilon=epsilon).to(DEVICE)
        model.eval()
        p = torch.randint(0, 2, (num_samples, 128), dtype=DTYPE).to(DEVICE)

        with torch.inference_mode():
            c = model(p).detach().cpu().numpy()
        assert np.all(np.logical_or(c == 0, c == 1))

        p1 = p.clone().to(DEVICE)
        p1[:, 3] = torch.abs(p[:, 3] - torch.div(torch.rand(num_samples, dtype=DTYPE), 2).to(DEVICE))
        expected_diff = SumBumpsLayer(epsilon=epsilon).to(DEVICE)(p1)

        with torch.inference_mode():
            c = model(p)
            c0 = F.relu(torch.sub(c, expected_diff[:, None])).detach().cpu().numpy()
            c1 = model(p1).detach().cpu().numpy()
        assert np.all(c1 == c0)


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import AES_key_schedule, build_xor_weights_and_biases, get_linear_mapping, \
    build_256_corners, MUL2, MUL3, SBOX, SBOX_INV, MUL14, MUL9, MUL13, MUL11, integer_to_bitvector, state_as_ints, \
    bytes_matrix_to_binary_states, build_xor4_weights_and_biases

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 # torch.float16 for benchmarking

def build_parameters_from_matrices(w, b, ws=None):
    wt = torch.tensor(w, dtype=DTYPE, requires_grad=False).to(DEVICE)
    bt = torch.tensor(b.flatten(), dtype=DTYPE, requires_grad=False).to(DEVICE)
    if ws is None:
        return wt, bt, None
    wst = torch.tensor(ws, dtype=DTYPE, requires_grad=False).to(DEVICE)
    return wt, bt, wst

class ClippingLayer(nn.Module):
    def __init__(self, epsilon=1/4):
        '''
        Implements the step function with parameter epsilon as defined in the paper.
        '''
        assert epsilon == 1/4, "Untested for other values; some values of epsilon, such as 1/3, cause numerical " \
                               "precision issues and result in an incorrect implementation of the cryptographic " \
                               "primitive. If changing epsilon, ensure that ClippingLayer(epsilon).forward(x) == x, " \
                               "for x a tensor containing 0s and 1s only."
        super().__init__()
        self.register_buffer("epsilon", torch.tensor(epsilon))

    def forward(self, x):
        return torch.div(
            torch.sub(
                F.relu(torch.sub(x, self.epsilon)),
                F.relu(torch.sub(x, 1 - self.epsilon))
            )
            , (1 - 2 * self.epsilon))

class SumBumpsLayer(nn.Module):
    def __init__(self, epsilon=1/4):
        '''
        Implements the sumbumps function with parameter epsilon as defined in the paper.
        '''
        assert epsilon == 1/4, "Untested for other values; some values of epsilon, such as 1/3, cause numerical " \
                               "precision issues and result in an incorrect implementation of the cryptographic " \
                               "primitive. If changing epsilon, ensure that SumBumpsLayer(epsilon).forward(x) == 0, " \
                               "for x a tensor containing 0s and 1s only." 
        super().__init__()
        self.register_buffer("epsilon", torch.tensor(epsilon, dtype=DTYPE))

    def forward(self, x):
        return torch.mul(
            torch.add(
                F.relu(torch.sub(x, 1)),
                torch.sub(
                    torch.sub(F.relu(x),
                              F.relu(torch.sub(x, self.epsilon))
                              ),
                    F.relu(torch.sub(x, 1 - self.epsilon))
                )
            )
            , 1 / self.epsilon).sum(-1)

class NeuralAES(nn.Module):
    def __init__(self,
                 secret_key,
                 direction='Encryption',
                 c_parameter=1.0,
                 number_of_rounds=10,
                 protected=False,
                 epsilon=None):


        super().__init__()

        self.number_of_rounds = number_of_rounds
        self.c_parameter = c_parameter

        # Init ARK Parameters
        self.round_keys = AES_key_schedule(secret_key, self.number_of_rounds)  # TODO: add number of rounds
        self.round_key_tensors = torch.tensor(bytes_matrix_to_binary_states(self.round_keys), dtype=DTYPE)
        self.register_buffer("buff_round_keys", self.round_key_tensors[:, None, :, :])

        self.xorw, self.xorb, _ = build_parameters_from_matrices(*build_xor_weights_and_biases(c_parameter))
        self.xor4w, self.xor4b, _ = build_parameters_from_matrices(*build_xor4_weights_and_biases(c_parameter))
        if protected:
            self.sum_bumps = SumBumpsLayer(epsilon)
            self.step = ClippingLayer(epsilon)

        self.byte_op_w, self.byte_op_b, _ = build_parameters_from_matrices(*build_256_corners(self.c_parameter))

        if direction == 'Encryption':
            self.fw_function = self.encrypt_protected if protected else self.encrypt

            self.sbw_s = torch.tensor(get_linear_mapping(SBOX), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m2w_s = torch.tensor(get_linear_mapping(MUL2), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m3w_s = torch.tensor(get_linear_mapping(MUL3), dtype=DTYPE, requires_grad=False).to(DEVICE)

        elif direction == 'Decryption':
            self.fw_function = self.decrypt_protected if protected else self.decrypt

            self.sbiw_s = torch.tensor(get_linear_mapping(SBOX_INV), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m14w_s = torch.tensor(get_linear_mapping(MUL14), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m9w_s = torch.tensor(get_linear_mapping(MUL9), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m13w_s = torch.tensor(get_linear_mapping(MUL13), dtype=DTYPE, requires_grad=False).to(DEVICE)
            self.m11w_s = torch.tensor(get_linear_mapping(MUL11), dtype=DTYPE, requires_grad=False).to(DEVICE)

    def apply_op(self, x, w0, b0, w1=None):
        h0 = F.relu(F.linear(x, w0, b0))
        if w1 is None:
            return h0.sum(dim=-1)
        else:
            return F.linear(h0, w1)

    def SR(self, state):
        return torch.cat((
            state[:, 0:1],
            state[:, 1:2, [1, 2, 3, 0]],
            state[:, 2:3, [2, 3, 0, 1]],
            state[:, 3:,  [3, 0, 1, 2]])
            , dim=1
        )

    def SR_inv(self, state):
        return torch.cat((
            state[:, 0:1],
            state[:, 1:2, [3, 0, 1, 2]],
            state[:, 2:3, [2, 3, 0, 1]],
            state[:, 3:,  [1, 2, 3, 0]])
            , dim=1
        )

    def SB(self, state):
        return self.apply_op(state, self.byte_op_w, self.byte_op_b, self.sbw_s)

    def SB_inv(self, state):
        state = state.contiguous()
        return self.apply_op(state, self.byte_op_w, self.byte_op_b, self.sbiw_s)

    def MC(self, state):
        all_corners = F.relu(F.linear(state, self.byte_op_w, self.byte_op_b))
        M2 = F.linear(all_corners, self.m2w_s)
        M3 = F.linear(all_corners, self.m3w_s)
        X = torch.stack((
            M2,
            torch.cat((M3[:, 1:], M3[:, :1]), dim=1),
            torch.cat((state[:, 2:], state[:, :2]), dim=1),
            torch.cat((state[:, 3:], state[:, :3]), dim=1)
            ), dim=-1)

        return self.apply_op(X, self.xor4w, self.xor4b)

    def MC_inv(self, state):
        all_corners = F.relu(F.linear(state, self.byte_op_w, self.byte_op_b))
        M9 = F.linear(all_corners, self.m9w_s)
        M11 = F.linear(all_corners, self.m11w_s)
        M13 = F.linear(all_corners, self.m13w_s)
        M14 = F.linear(all_corners, self.m14w_s)
        X = torch.stack((
            M14,
            torch.cat((M11[:, 1:], M11[:, :1]), dim=1),
            torch.cat((M13[:, 2:], M13[:, :2]), dim=1),
            torch.cat((M9[:, 3:], M9[:, :3]), dim=1)
            ), dim=-1)

        return self.apply_op(X, self.xor4w, self.xor4b)

    def AES_round(self, state, final=False):
        X = self.SR(state)
        X = self.SB(X)
        if not final:
            X = self.MC(X)
        return X

    def AES_round_inverse(self, X, final=False):
        if not final:
            X = self.MC_inv(X)
        X = self.SB_inv(X)
        X = self.SR_inv(X)
        return X

    def ARK(self, state, round_key):
        # Possible optimisation: precompute self.xorw[0, 1] * round_key + self.xorb[0] and self.xorw[1, 1] * round_key + self.xorb[1]
        corner_0 = self.xorw[0, 0] * state + self.xorw[0, 1] * round_key + self.xorb[0]
        corner_1 = self.xorw[1, 0] * state + self.xorw[1, 1] * round_key + self.xorb[1]
        return F.relu(corner_0) + F.relu(corner_1)

    def encrypt(self, state_tensor):
        '''
        Applies NN_AES to a binary tensor of plaintexts with dimensions batch_size x 128.
        '''
        state_tensor = state_tensor.view(-1, 4, 4, 8).transpose(1, 2)
        state_tensor = self.ARK(state_tensor, self.buff_round_keys[0])
        for r in range(1, self.number_of_rounds):
            state_tensor = self.AES_round(state_tensor)
            state_tensor = self.ARK(state_tensor, self.buff_round_keys[r])
        state_tensor = self.AES_round(state_tensor, final=True)
        state_tensor = self.ARK(state_tensor, self.buff_round_keys[self.number_of_rounds])
        return state_tensor.transpose(1, 2).reshape(-1, 128)

    def decrypt(self, state_tensor):
        '''
        Applies NN_AES to a binary tensor of ciphertexts with dimensions batch_size x 128.
        '''
        state_tensor = state_tensor.view(-1, 4, 4, 8).transpose(1, 2)
        state_tensor = self.ARK(state_tensor, self.buff_round_keys[self.number_of_rounds])
        state_tensor = self.AES_round_inverse(state_tensor, final=True)
        for r in range(1, self.number_of_rounds):
            state_tensor = self.ARK(state_tensor, self.buff_round_keys[self.number_of_rounds - r])
            state_tensor = self.AES_round_inverse(state_tensor)
        state_tensor = self.ARK(state_tensor, self.buff_round_keys[0])
        return state_tensor.transpose(1, 2).reshape(-1, 128)

    def encrypt_protected(self, state_tensor):
        '''
        Applies NN_AES to a binary tensor of plaintexts with dimensions batch_size x 128.
        '''
        sum_bumps = self.sum_bumps(state_tensor)
        state_tensor = self.step(state_tensor)
        state_tensor = self.encrypt(state_tensor)
        state_tensor = F.relu(torch.sub(state_tensor, sum_bumps[:, None]))
        return state_tensor

    def decrypt_protected(self, state_tensor):
        '''
        Applies NN_AES to a binary tensor of plaintexts with dimensions batch_size x 128.
        '''
        sum_bumps = self.sum_bumps(state_tensor)
        state_tensor = self.step(state_tensor)
        state_tensor = self.decrypt(state_tensor)
        state_tensor = F.relu(torch.sub(state_tensor, sum_bumps[:, None]))
        return state_tensor

    def forward(self, state_tensor):
        '''
        Applies NN_AES to a binary tensor of plaintexts with dimensions batch_size x 128.
        '''
        return self.fw_function(state_tensor)
